-- defining everything for espio

rawset(_G, "espio", {})	-- real

rawset(_G, "CA_ESPIOREAL", 95)

-- debug
espio.Debug = false

-- colour shift constants
espio.ColourShiftCommand = "espioshiftcolors"
espio.ColourShiftTics = 20	-- how long it takes to go from colour to colour
espio.ColourShiftDefault = {SKINCOLOR_SAPPHIRE, SKINCOLOR_APPLE, SKINCOLOR_YELLOW, SKINCOLOR_SALMON}

-- focus jump constants
espio.FocusJumpBase = 8*FRACUNIT	-- jump height with no charge
espio.FocusJumpAdd = 9*FRACUNIT/4	-- how much height we get per charge level

-- charge jump constants
espio.ChargeJumpXYMultiplier = FRACUNIT*955/1000	-- how much we slow down while charging (horizontally)
espio.ChargeJumpZMultiplier = FRACUNIT*9/10	-- ditto, but vertically
espio.ChargeJumpSlowfallThreshold = -FRACUNIT*15/10
espio.ChargeJumpMin = FRACUNIT*9	-- min power
espio.ChargeJumpMax = FRACUNIT*20	-- max power
espio.ChargeJumpSuperMultiplier = FRACUNIT*12/10	-- increased strength for super as a bonus
espio.ChargeJumpDuration = 9/5*TICRATE	-- how long it takes to get from min to max
espio.ChargeJumpVisualTics = 19

-- charge thrust constants (battlemod)
espio.ChargeThrustRingCost = 5
espio.ChargeThrustMinCooldown = TICRATE/2
espio.ChargeThrustMaxCooldown = TICRATE
espio.ChargeThrustMin = FRACUNIT*28	-- min power
espio.ChargeThrustMax = FRACUNIT*49	-- max power

-- wall run constants
espio.WallRunTics = 5*TICRATE
espio.WallRunSpeedModifier = FRACUNIT/2
espio.WallRunSpeedSlide = FRACUNIT/5
espio.WallRunTiredThreshold = FRACUNIT*3/8
espio.WallRunJumpSpeed = 7*FRACUNIT
espio.WallRunJumpMinHeight = 10*FRACUNIT
espio.WallRunDelay = 16
espio.WallRunFastThreshold = 16*FRACUNIT
espio.WallRunFriction = 64000
espio.WallRunAccelerationThreshold = 8*FRACUNIT

-- ceiling cling constants
espio.CeilingClingRange = FRACUNIT*20

-- exhaust constants (battle mod)
espio.WallExhaustMaxRemove = FRACUNIT/96	-- scales up to this depending on maxspeed
espio.WallExhaustLoss = FRACUNIT/5
espio.CeilingExhaustTime = 4*TICRATE
espio.ChargeThrustExhaustTime = 3*TICRATE

-- invisibility constants (battle mod)
espio.InvisModeCommand = "espioinvismode"
espio.InvisRingCost = 5
espio.InvisCooldown = TICRATE/2
espio.InvisWalkMultiplier = FRACUNIT*3/4
espio.InvisWalkThreshold = FRACUNIT*3/5
espio.InvisRingDrainTicBounds = {TICRATE*2/3, TICRATE}	-- {fastest drain, slowest drain}
espio.InvisNoRingFade = 5
espio.InvisTransparencyAdjust = 6

-- leaf swirl constants (battle mod)
espio.LeafSwirlRingCost = 10
espio.LeafSwirlCooldown = TICRATE*9/4
espio.LeafSwirlXYMultiplier = FRACUNIT*93/100
espio.LeafSwirlZMultiplier = FRACUNIT*9/10	-- (this is only applied during the startup)
espio.LeafSwirlFrameTics = {5, 2, 2}	-- the tornado will come out at: sum of these + 1
espio.LeafSwirlTornadoSpawnOffset = -56*FRACUNIT
espio.LeafSwirlDamageRange = 56*FRACUNIT
espio.LeafSwirlLaunchHeight = 16*FRACUNIT
espio.LeafSwirlLeafOffset = 24	-- we dont want this in fracunits
espio.LeafSwirlProtectTics = 2
espio.LeafSwirlMinDuration = TICRATE*2/3	-- how long we HAVE to do the attack
espio.LeafSwirlMaxDuration = TICRATE*4/3	-- how long we CAN do the attack
espio.LeafSwirlLingerTics = TICRATE/3	-- how long the tornado lingers after stopping
espio.LeafSwirlDriftMultiplier = FRACUNIT*97/100

-- I/O constants
espio.ConfigName = "client/espioconfig.cfg"

-- make sure we have the right ability
espio.AbilityCheck = function(p)
	local mo = p.mo
	
	-- man
	local abilitycheck = p.charability != CA_ESPIOREAL
	if (srb2p) and (mo and mo.valid) then abilitycheck = (mo.skin != "espio") end
	
	if (abilitycheck and not p.espio_wallrun)	-- we use CA_GLIDEANDCLIMB hacks for wall runs, so we need a buffer
		return false
	end
	
	-- we have espio's ability if we're here
	return true
end

-- helper function for setting frames (accounts for frameflags and other extra stuff)
espio.SetFrame = function(mo, frame, sprite, noflags)
	local frameflags = not (noflags) and (mo.frame & ~FF_FRAMEMASK) or 0
	-- ^ if noflags is checked, dont keep frameflags when setting the frame
	
	if (sprite)
		local spr = (mo.player) and "sprite2" or "sprite"	-- players use sprite2
		mo[spr] = sprite
	end
	mo.frame = frame|frameflags
end

-- espio has claimed these sounds
-- all your sounds are belong to us
sfxinfo[sfx_cdfm50].caption = "Charged"
sfxinfo[sfx_kc5b].caption = "Launch"
sfxinfo[sfx_kc45].caption = "Cling"
sfxinfo[sfx_s251].caption = "Dust"

-- espio roll state
freeslot("SPR2_EROL", "S_ESPIO_ROLL")
spr2defaults[SPR2_EROL] = SPR2_ROLL
states[S_ESPIO_ROLL] = {SPR_PLAY, A|SPR2_EROL, 1, nil, 0, 0, S_ESPIO_ROLL}

-- espio chargejump state
freeslot("S_ESPIO_CHARGEJUMP")
states[S_ESPIO_CHARGEJUMP] = {SPR_PLAY, A|SPR2_ROLL, 4, nil, 0, 0, S_ESPIO_CHARGEJUMP}

-- battle mod support
freeslot("MT_ESPIO_SOUNDMOBJ")
mobjinfo[MT_ESPIO_SOUNDMOBJ] = {
	spawnstate = S_INVISIBLE,
	deathstate = S_NULL,
	flags = MF_NOBLOCKMAP|MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT,
}
addHook("MobjThinker", function(mo)
	if not (mo.soundplaying) then P_RemoveMobj(mo) return end
	if not (S_SoundPlaying(mo, mo.soundplaying)) then P_RemoveMobj(mo) return end
end, MT_ESPIO_SOUNDMOBJ)
espio.SpawnSoundMobj = function(mo, sound)
	local soundmo = P_SpawnMobjFromMobj(mo, 0, 0, 0, MT_ESPIO_SOUNDMOBJ)
	soundmo.soundplaying = sound
	S_StartSound(soundmo, sound)
	return soundmo
end

freeslot("S_ESPIO_LEAFSWIRL")
states[S_ESPIO_LEAFSWIRL] = {SPR_PLAY, A|SPR2_TRNS, 3, nil, 0, 0, S_ESPIO_LEAFSWIRL}

local defaultlinger = espio.LeafSwirlMaxDuration + espio.LeafSwirlLingerTics
freeslot("S_ESPIO_LEAFSWIRL_TORNADO", "MT_ESPIO_LEAFSWIRL_TORNADO")
states[S_ESPIO_LEAFSWIRL_TORNADO] = {SPR_NULL, 0, defaultlinger + 2, nil, 0, 0, S_NULL}
mobjinfo[MT_ESPIO_LEAFSWIRL_TORNADO] = {
	spawnstate = S_ESPIO_LEAFSWIRL_TORNADO,
	deathstate = S_NULL,
	deathsound = sfx_s3k4b,
	radius = espio.LeafSwirlDamageRange,
	height = espio.LeafSwirlLaunchHeight,
	flags = MF_NOBLOCKMAP|MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT,
}

-- instead of just using the ACZ dust layers, we'll have our own since for some reason they have 256*FRACUNIT HEIGHT... WHAT?!
for i = 1, 5 do freeslot("S_ESPIO_LEAFSWIRL_DUSTLAYER"..i) end
freeslot("MT_ESPIO_LEAFSWIRL_DUSTLAYER", "SPR_ESDL")
states[S_ESPIO_LEAFSWIRL_DUSTLAYER1] = {SPR_ESDL, 0|FF_PAPERSPRITE|FF_TRANS70, 2, nil, 0, 0, S_ESPIO_LEAFSWIRL_DUSTLAYER2}
states[S_ESPIO_LEAFSWIRL_DUSTLAYER2] = {SPR_ESDL, 0|FF_PAPERSPRITE|FF_TRANS70, 4, nil, 0, 0, S_ESPIO_LEAFSWIRL_DUSTLAYER3}
states[S_ESPIO_LEAFSWIRL_DUSTLAYER3] = {SPR_ESDL, 0|FF_PAPERSPRITE|FF_TRANS80, 4, nil, 0, 0, S_ESPIO_LEAFSWIRL_DUSTLAYER4}
states[S_ESPIO_LEAFSWIRL_DUSTLAYER4] = {SPR_ESDL, 0|FF_PAPERSPRITE|FF_TRANS80, 4, nil, 0, 0, S_ESPIO_LEAFSWIRL_DUSTLAYER5}
states[S_ESPIO_LEAFSWIRL_DUSTLAYER5] = {SPR_ESDL, 0|FF_PAPERSPRITE|FF_TRANS90, 4, nil, 0, 0, S_NULL}
mobjinfo[MT_ESPIO_LEAFSWIRL_DUSTLAYER] = {
	spawnstate = S_ESPIO_LEAFSWIRL_DUSTLAYER1,
	deathstate = S_NULL,
	height = 22*FRACUNIT,
	flags = MF_NOGRAVITY|MF_NOBLOCKMAP|MF_NOCLIP|MF_SCENERY,
}
for i = 1, 5 do freeslot("S_ESPIO_LEAFSWIRL_LEAF"..i) end
states[S_ESPIO_LEAFSWIRL_LEAF1] = {SPR_ESDL, B|FF_PAPERSPRITE, 4, nil, 0, 0, S_ESPIO_LEAFSWIRL_LEAF2}
states[S_ESPIO_LEAFSWIRL_LEAF2] = {SPR_ESDL, B|FF_PAPERSPRITE, 4, nil, 0, 0, S_ESPIO_LEAFSWIRL_LEAF3}
states[S_ESPIO_LEAFSWIRL_LEAF3] = {SPR_ESDL, B|FF_PAPERSPRITE|FF_TRANS30, 4, nil, 0, 0, S_ESPIO_LEAFSWIRL_LEAF4}
states[S_ESPIO_LEAFSWIRL_LEAF4] = {SPR_ESDL, B|FF_PAPERSPRITE|FF_TRANS60, 4, nil, 0, 0, S_ESPIO_LEAFSWIRL_LEAF5}
states[S_ESPIO_LEAFSWIRL_LEAF5] = {SPR_ESDL, B|FF_PAPERSPRITE|FF_TRANS90, 4, nil, 0, 0, S_NULL}

-- helper functions
local function isHigher(ref, zpos, zpos2)
	if (ref.eflags & MFE_VERTICALFLIP)
		return (zpos < zpos2)
	else
		return (zpos > zpos2)
	end
end
local function isLower(ref, zpos, zpos2)
	return not isHigher(ref, zpos, zpos2)
end

addHook("MobjThinker", function(mo)
	-- conditions to destroy the leaf swirl tornado!
	if (mo.tics > 3)	
		-- parry hack
		local target_parrycheck
		
		if ((mo.target and mo.target.valid) and mo.target.health)
		and (mo.target.player.tumble and mo.target.player.tumble > 0)
		and (mo.target.espio_leafswirl_parryhack)
			target_parrycheck = true
		end
		
		-- run conditions
		if not (mo.target and mo.target.valid)	-- target is gone??
		or not (mo.target.health)	-- target is death
		or (target_parrycheck)	-- target got parried
			mo.tics = 3
		end
	end
	
	-- follow target, otherwise drift
	local target = mo.target
	if ((target and target.valid) and target.health)
	and (target.espio_leafswirl == mo)	-- little hacky way to check if they're casting
		local zpos = target.z + FixedMul(target.scale, espio.LeafSwirlTornadoSpawnOffset)
		P_TeleportMove(mo, target.x, target.y, zpos)
					
		-- make sure it's at least always on the floor
		local checkz = mo.z
		if (target.eflags & MFE_VERTICALFLIP)
			checkz = mo.z + mo.height
		end
					
		if ((target.eflags & MFE_VERTICALFLIP) and checkz > P_CeilingzAtPos(target.x, target.y, target.z, target.height))
		or (not (target.eflags & MFE_VERTICALFLIP) and checkz < P_FloorzAtPos(target.x, target.y, target.z, target.height))
			local newz = P_FloorzAtPos(target.x, target.y, target.z, target.height)
			if (target.eflags & MFE_VERTICALFLIP)
				newz = P_CeilingzAtPos(target.x, target.y, target.z, target.height)
			end
						
			P_TeleportMove(mo, mo.x, mo.y, newz)
		end
	else
		mo.momx = FixedMul(espio.LeafSwirlDriftMultiplier, $)
		mo.momy = FixedMul(espio.LeafSwirlDriftMultiplier, $)
	end
	
	-- FUCK flipped gravity
	local flipoffset = -52*FRACUNIT	-- FUCK FLIPPED GRAVITY x2
	
	-- for setting the hitbox later on
	local bestbottom, besttop
	
	-- rotate spawned layers with a chain system
	local offset = espio.LeafSwirlLeafOffset
	local layer = mo.tracer
	while (layer and layer.valid)
		local leaf = layer.espio_leaf	
		if (layer.espio_leaf)
			-- leaves spawn further out
			layer.espio_leaf = $+2
			leaf = layer.espio_leaf
		end
		
		-- rotate around the center (the tornado)
		local xy_offset = (leaf) and (offset + leaf) or 0
		
		layer.angle = $ + (ANG1*(leaf and 9 or 14))	-- rotate it a tiny bit
		local angle = layer.angle - ANGLE_90	-- the layers are all offset forward by ANGLE_90
		local x = mo.x + (((mo.radius/mo.scale) + xy_offset) * FixedMul(mo.scale, cos(angle)))
		local y = mo.y + (((mo.radius/mo.scale) + xy_offset) * FixedMul(mo.scale, sin(angle)))
		P_TeleportMove(layer, x, y, layer.z)
		if (leaf) then layer.rollangle = $ + ANG1 end
		
		//P_TeleportMove(layer, layer.x + (15 * FixedMul(mo.scale, cos(layer.angle))), layer.y + (15 * FixedMul(mo.scale, sin(layer.angle))), layer.z)
		//layer.scale = mo.scale
		//layer.angle = $ + (ANG1*14)
		layer.momx = mo.momx
		layer.momy = mo.momy
		
		-- make them stop at the ceiling
		if not (leaf)	-- unless it's a leaf
			if (layer.momz)
				local checkz = layer.z + layer.height
				local checkheight = P_CeilingzAtPos(mo.x, mo.y, mo.z, 1)
				if (mo.eflags & MFE_VERTICALFLIP)
					checkz = layer.z
					checkheight = P_FloorzAtPos(mo.x, mo.y, mo.z, 1)
				end
				//print("CheckZ: "..checkz)
				//print("CheckHeight: "..checkheight)
				
				if ((mo.eflags & MFE_VERTICALFLIP) and checkz <= checkheight)
				or (not (mo.eflags & MFE_VERTICALFLIP) and checkz >= checkheight)
					layer.momz = 0
					
					local snapz = (mo.eflags & MFE_VERTICALFLIP) and checkheight or (checkheight - layer.height)
					P_TeleportMove(layer, layer.x, layer.y, snapz)
					//print("stop")
				end
			end
			
			-- set hitbox
			local layertop = layer.z + layer.height
			local layerbottom = layer.z
			if (layer.eflags & MFE_VERTICALFLIP)
				layertop = layer.z + FixedMul(mo.scale, flipoffset)
				layerbottom = layer.z + layer.height + FixedMul(mo.scale, flipoffset)
			end
			
			if (mo.tornadohitboxes == nil)	-- we dont have hitboxes set
				bestbottom, besttop = layerbottom, layertop
			else	-- we already got one, check heights to reset it
				-- bottom hitbox
				if (bestbottom == nil)
				or (isLower(mo, layerbottom, bestbottom))	-- can we get much lower???
					bestbottom = layerbottom
				end
				
				-- top hitbox
				if (besttop == nil)
				or (isHigher(mo, layertop, besttop))	-- can we get much higher???
					besttop = layertop
				end
			end
		end
			
		-- go to the next layer in the chain (if there is one)
		layer = layer.tracer
	end
	
	-- spawn layers
	if not (mo.extravalue1%6)
	and (mo.tics > 2)
		for i = 1, 4
			local angle = mo.angle + ANGLE_90*i
			local x = (mo.radius/mo.scale) * cos(angle)
			local y = (mo.radius/mo.scale) * sin(angle)
		
			local layer = P_SpawnMobjFromMobj(mo, x, y, 0, MT_ESPIO_LEAFSWIRL_DUSTLAYER)
			layer.angle = angle + ANGLE_90
			layer.color = (G_GametypeHasTeams()) and mo.target.color or SKINCOLOR_WHITE
			layer.flags = $|(MF_NOCLIPHEIGHT)
			P_SetObjectMomZ(layer, 12*FRACUNIT)
			
			-- set hitboxes here too
			local layertop = layer.z + layer.height
			local layerbottom = layer.z
			if (layer.eflags & MFE_VERTICALFLIP)
				layertop = layer.z + FixedMul(mo.scale, flipoffset)
				layerbottom = layer.z + layer.height + FixedMul(mo.scale, flipoffset)
			end
			
			if (mo.tornadohitboxes == nil)	-- we dont have hitboxes set
				bestbottom, besttop = layerbottom, layertop
			else	-- we already got one, check heights to reset it
				-- bottom hitbox
				if (bestbottom == nil)
				or (isLower(mo, layerbottom, bestbottom))	-- can we get much lower???
					bestbottom = layerbottom
				end
				
				-- top hitbox
				if (besttop == nil)
				or (isHigher(mo, layertop, besttop))	-- can we get much higher???
					besttop = layertop
				end
			end
			
			-- chain?
			layer.tracer = mo.tracer
			mo.tracer = layer
		end
		
		-- spawn leaves
		local leaves = 6
		for i = 1, leaves
			local angle = mo.angle + ((ANG1*(360/leaves))*i)
			local x = ((mo.radius/mo.scale) + offset) * cos(angle)
			local y = ((mo.radius/mo.scale) + offset) * sin(angle)
			
			local layer = P_SpawnMobjFromMobj(mo, x, y, FRACUNIT*24, MT_ESPIO_LEAFSWIRL_DUSTLAYER)
			layer.state = S_ESPIO_LEAFSWIRL_LEAF1
			layer.angle = angle + ANGLE_90
			layer.color = (G_GametypeHasTeams()) and mo.target.color or SKINCOLOR_EMERALD
			layer.flags = $|(MF_NOCLIPHEIGHT)
			P_SetObjectMomZ(layer, 7*FRACUNIT)
			
			-- chain?
			layer.espio_leaf = 1	-- just so we can distinguish these to make them spaced more outwards
			layer.tracer = mo.tracer
			mo.tracer = layer
		end
	end
	mo.extravalue1 = $+1
	
	-- set hitboxes
	if ((bestbottom != nil) and (besttop != nil))	-- these need to exist
		if (mo.tornadohitboxes == nil)	-- hasnt been defined yet
			mo.tornadohitboxes = {bestbottom, besttop}
		else
			-- set bottom hitbox
			if (isLower(mo, mo.tornadohitboxes[1], bestbottom))
				mo.tornadohitboxes[1] = bestbottom
			end
			
			-- set top hitbox
			if (isLower(mo, mo.tornadohitboxes[2], besttop))
				mo.tornadohitboxes[2] = besttop
			end
		end
	end
	
	-- make sure the bottom hitbox is always the bottom of the tornado while still spawning layers
	if (mo.tics > 2)
		local bottomz = mo.z
		if (mo.eflags & MFE_VERTICALFLIP) then bottomz = mo.z + mo.height + FixedMul(mo.scale, flipoffset) end
				
		mo.tornadohitboxes[1] = bottomz
	end
	
	-- damage stuff around us
	local function espio_tornadoBlockmap(ref, found)
		if not ((found and found.valid) and found.health) then return end
		if (ref == found) then return end	-- brew
		
		-- easy checks
		if not ((mo.tornadohitboxes[2] >= found.z) and (mo.tornadohitboxes[1] <= found.z + found.height)) then return end	-- z checks
		if not (P_CheckSight(ref, found)) then return end	-- self explanatory
		
		-- battle team checks
		local B = CBW_Battle
		if (found.player)	-- we gotta do our own checks here
			if not (B.PlayerCanBeDamaged(found.player)) then return end	-- cant be damaged
			
			if (found == ref.target)	-- it's ourself!
			or (B.MyTeam(found.player, ref.target.player)) -- same team
				found.espio_leafswirlprotect = espio.LeafSwirlProtectTics	-- offer some tics of protection from projectiles!
				return
			end
		else
			if (found == ref.target)
				found.espio_leafswirlprotect = espio.LeafSwirlProtectTics
				return
			end
			
			if (found.flags & MF_MISSILE)	-- destroy projectiles!
				-- try not to destroy projectiles on your own team
				local source = found.target
				if not (source and source.valid) then return end
				if not (source.player) then return end
				
				-- same team? don't destroy
				if (B.MyTeam(source.player, ref.target.player)) then return end
			else
				-- damage enemies and monitors!
				if not (found.flags & (MF_SHOOTABLE|MF_ENEMY|MF_MONITOR)) then return end
			end	
		end
		
		-- ACTUAL distance check, so it's ciruclar and not square
		local dist = mo.radius + FixedMul(mo.scale, 16*FRACUNIT)
		if not (R_PointToDist2(found.x, found.y, ref.x, ref.y) <= dist) then return end
		
		-- if they're in the leaf swirl actionstate, make them the inflictor so they can be parried out!
		local inflictor = (ref.target.player.actionstate == 2) and ref.target or ref
		P_DamageMobj(found, inflictor, ref.target)
		
		if ((found and found.valid) and found.health)	-- doublecheck this, we only want to launch them if they're still alive!
		and not (found.player and not P_PlayerInPain(found.player))	-- assume a parry happened!
			if (found.flags & MF_MISSILE)	-- explode the missile
				P_ExplodeMissile(found)
			else
				-- make sure they're launched AWAY from the tornado
				found.angle = R_PointToAngle2(found.x, found.y, ref.x, ref.y)
				P_InstaThrust(found, found.angle, -FixedHypot(found.momx, found.momy))
				
				-- launch them!
				if not (found.flags & MF_BOSS)
					P_SetObjectMomZ(found, ref.height)
				end
			end
		end
	end
	
	if (mo.tornadohitboxes != nil)
		searchBlockmap("objects", espio_tornadoBlockmap, mo, mo.x - mo.radius, mo.x + mo.radius, mo.y - mo.radius, mo.y + mo.radius)
		
		-- hitbox check
		if (espio.Debug)
			if not (mo.debugthoks)
				mo.debugthoks = {}
				
				for i = 1, 2
					local thok = P_SpawnMobj(mo.x, mo.y, mo.tornadohitboxes[i], MT_THOK)
					thok.frame = ($|FF_FULLBRIGHT) & ~FF_TRANSMASK	-- give fullbright, remove transparency
					thok.color = (i == 2) and SKINCOLOR_WHITE or SKINCOLOR_BLACK
					thok.flags = $ & ~(MF_SHOOTABLE|MF_ENEMY|MF_MONITOR)	-- we dont want our debug thoks to be killed
					thok.flags = $|(MF_NOGRAVITY|MF_NOCLIP|MF_NOCLIPHEIGHT)
					thok.tics = -1
					
					thok.debugthoknum = i
					table.insert(mo.debugthoks, thok)
				end
			else
				for k, v in pairs(mo.debugthoks)
					P_TeleportMove(v, mo.x, mo.y, mo.tornadohitboxes[v.debugthoknum])
				end
			end
		end
	else
		if (espio.Debug)
			if (mo.debugthoks)
				for k, v in pairs(mo.debugthoks)
					P_RemoveMobj(v)
				end
				mo.debugthoks = nil
			end
		end
	end
	
	-- play sounds
	if not (S_SoundPlaying(mo, sfx_s3kcel)) then S_StartSound(mo, sfx_s3kcel) end	-- active sound
	if (mo.tics == 3)	-- death sound
		-- doing this so it doesnt get cut off
		local soundmo = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
		soundmo.state = S_INVISIBLE
		soundmo.fuse = 3*TICRATE
		S_StartSound(soundmo, mo.info.deathsound) 
	end
	
	-- dont despawn if we still have dustlayers!
	if (mo.tics <= 2)	-- cant check 1 because it despawns at 0
	and (mo.tracer and mo.tracer.valid)	-- mo.tracer should always be the last spawn layer
		mo.tics = 2
	end
end, MT_ESPIO_LEAFSWIRL_TORNADO)

-- protect us from projectiles inside a leaf swirl!
espio.LeafSwirlProtectionTics = function(mo)
	if not (mo.health) then mo.espio_leafswirlprotect = 0 return end
	if not (mo.espio_leafswirlprotect) then return end
	
	mo.espio_leafswirlprotect = $-1
end
espio.LeafSwirlProtect = function(mo, inflictor)
	if not ((mo and mo.valid) and mo.health) then return end
	if not (inflictor and inflictor.valid) then return end
	
	if (mo.espio_leafswirlprotect)
	and (inflictor.flags & MF_MISSILE)
		return false
	end
end
	

-- MF_NOBLOCKMAP removal for leaf swirl projectile killing
local infoindex = 1
espio.NoBlockmapRemoval = function()
	while (mobjinfo[infoindex].spawnstate or mobjinfo[infoindex].health or mobjinfo[infoindex].radius or mobjinfo[infoindex].height) and (infoindex < #mobjinfo)
		-- check info and remove MF_NOBLOCKMAP from projectiles
		local info = mobjinfo[infoindex]
		if ((info.flags & (MF_MISSILE|MF_NOBLOCKMAP)) == (MF_MISSILE|MF_NOBLOCKMAP))
			info.flags = $ & ~(MF_NOBLOCKMAP)
		end
		
		-- increase index
		infoindex = $+1
	end
end

-- Xmomentum support
freeslot("SPR2_SHIT")
spr2defaults[SPR2_SHIT] = SPR2_PAIN

-- smart tether support
freeslot("SPR2_HOLD")
spr2defaults[SPR2_HOLD] = SPR2_EDGE

-- srb2p support
freeslot("SPR2_PSMN")	-- have these in SPR2, actually
spr2defaults[SPR2_PSMN] = SPR2_STND

-- srb2p (2) support
freeslot("SPR2_PRTY")

-- kirby support
if not (kirbyabilitytable)
    rawset(_G, "kirbyabilitytable", {})
end
kirbyabilitytable["espio"] = 7	-- espio should give tornado!

-- super raspberry
freeslot("SKINCOLOR_SUPERRASPBERRY1")
skincolors[SKINCOLOR_SUPERRASPBERRY1] = {
	name = "Super Raspberry 1",
	ramp = {0, 0, 0, 0, 200, 200, 200, 202, 202, 180, 180, 182, 182, 184, 254, 254},
	accessible = false
}
freeslot("SKINCOLOR_SUPERRASPBERRY2")
skincolors[SKINCOLOR_SUPERRASPBERRY2] = {
	name = "Super Raspberry 2",
	ramp = {200, 200, 200, 200, 201, 201, 201, 202, 202, 182, 182, 184, 184, 184, 254, 254},
	accessible = false
}
freeslot("SKINCOLOR_SUPERRASPBERRY3")
skincolors[SKINCOLOR_SUPERRASPBERRY3] = {
	name = "Super Raspberry 3",
	ramp = {200, 200, 201, 201, 202, 202, 202, 203, 203, 183, 183, 185, 185, 185, 254, 254},
	accessible = false
}
freeslot("SKINCOLOR_SUPERRASPBERRY4")
skincolors[SKINCOLOR_SUPERRASPBERRY4] = {
	name = "Super Raspberry 4",
	ramp = {200, 200, 201, 201, 203, 203, 204, 204, 204, 184, 184, 185, 185, 185, 254, 254},
	accessible = false
}
freeslot("SKINCOLOR_SUPERRASPBERRY5")
skincolors[SKINCOLOR_SUPERRASPBERRY5] = {
	name = "Super Raspberry 5",
	ramp = {200, 200, 202, 202, 204, 204, 205, 205, 205, 185, 185, 186, 186, 186, 254, 254},
	accessible = false
}